# logger.py -- Put log messages somewhere.
# Gareth Rees, Ravenbrook Limited, 2000-10-16.
# $Id: //info.ravenbrook.com/project/p4dti/version/1.0/code/replicator/logger.py#3 $
#
# This file is copyright (c) 2001 Perforce Software, Inc.  All rights  
# reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1.  Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
# 
# 2.  Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDERS AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

import string
import sys
import time
import types

class logger:
    def date_message(self, id, format, arguments = ()):
        message = self.message(id, format, arguments)
        date = time.strftime('%Y-%m-%d %H:%M:%S UTC', time.gmtime(time.time()))
        return string.expandtabs(("%s\t%s") % (date, message))
    # message(id, format, arguments).  Format log message and return the
    # message as a string.
    def message(self, id, format, arguments = ()):
        assert isinstance(id, types.StringType)
        assert isinstance(format, types.StringType)
        check_digit = self.check_digit(id)
        if type(arguments) != types.TupleType:
            arguments = (arguments,)
        return string.expandtabs(("%s%s\t" + format)
                                 % ((id, check_digit) + arguments))

    # Compute the check digit for the given message id.  Message ids look like
    # P-NC, where P is the project name, N is the id number and C is the check
    # digit, computed by this function.  See
    # <http://info.ravenbrook.com/mail/2000/10/16/10-50-49/0.txt> for the
    # design decision.
    def check_digit(self, id):
        assert isinstance(id, types.StringType)
        sum = 0
        i = 0
        try:
            while 1:
                i = i - 1
                j = string.index(string.digits, id[i])
                sum = sum - i * j
        except (ValueError, IndexError):
            return "0123456789X"[sum % 11]
    
    # log(id, format, arguments, priority).  Log the message.

class file_logger(logger):
    file = None
    def __init__(self, file = sys.stdout):
        self.file = file
    def log(self, id, format, arguments = (), priority = None):
        self.file.write(self.date_message(id, format, arguments)+'\n')
        self.file.flush()

# log to syslog on Unix.  On other systems, log to /dev/null

class sys_logger(logger):
    def syslog(self, text):
        pass
    def __init__(self):
        import os
        if os.name == 'posix':
            import syslog
            self.syslog = syslog.syslog
            syslog.openlog('p4dti', syslog.LOG_PID, syslog.LOG_DAEMON)
    def log(self, id, format, arguments = (), priority = None):
        self.syslog(self.message(id, format, arguments))

class multi_logger(logger):
    loggers = []
    def __init__(self, loggers = []):
        self.loggers = loggers
    def log(self, id, format, arguments = (), priority = None):
        for l in self.loggers:
            l.log(id, format, arguments, priority)

# B. Document History
# 
# 2000-10-16 GDR Created.
# 
# 2000-11-30 GDR Added some type checking.
#
# 2001-01-19 NB  Added sys_logger.
#
# 2001-01-26 NB  Removed extra \n.
#
# 2001-03-01 NB  Fix for job000237.
#
# 2001-03-02  RB    Transferred copyright to Perforce under their license.
