# migrate_teamtrack.py -- Migrate Perforce jobs to TeamTrack.
#
# Gareth Rees, Ravenbrook Limited, 2000-10-19.
#
# $Id: //info.ravenbrook.com/project/p4dti/version/1.0/code/replicator/migrate_teamtrack.py#6 $
#
# 1. Introduction
# 
# This file contains the outline of a program that can convert Perforce jobs
# into TeamTrack cases and submit them to TeamTrack.  Since I don't know what
# fields you have in your Perforce jobs or how you plan to convert them to
# fields in TeamTrack, I can't prepare a complete solution for you.  Instead,
# you'll have to go through this file, read the comments and fill in the
# details.
#
# Each instance where you need to think and possibly edit the code is marked
# with three asterisks (***).
#
# Run this script from the command line by switching to the directory
# containing this file, then run the command "python migrate_teamtrack.py".
# 
# You should also read section 6.2 of the P4DTI Administrator's Guide.
#
# ---
#
# This file is copyright (c) 2001 Perforce Software, Inc.  All rights  
# reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1.  Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
# 
# 2.  Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDERS AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

from config import *
import configure_teamtrack
import dt_teamtrack
import p4
import string
import teamtrack

# Generate the configuration for the dt_teamtrack module.
teamtrack_config, _, _, _ = \
    configure_teamtrack.configuration \
    (
     administrator_address = administrator_address,
     changelist_url = changelist_url,
     closed_state = closed_state,
     log_file = log_file,
     p4_client_executable = p4_client_executable,
     p4_password = p4_password,
     p4_port = p4_port,
     p4_server_description = p4_server_description,
     p4_user = p4_user,
     poll_period = poll_period,
     replicate_p = replicate_p,
     replicator_address = replicator_address,
     replicated_fields = replicated_fields,
     smtp_server = smtp_server,
     start_date = start_date,
     teamtrack_password = teamtrack_password,
     teamtrack_server = teamtrack_server,
     teamtrack_user = teamtrack_user,
     verbose = verbose,
     )

# Construct a dt_teamtrack instance.
dt = dt_teamtrack.dt_teamtrack(rid, sid, teamtrack_config)

# Setup of TeamTrack for the P4DTI by adding fields to the CASES table.
dt.init()

# convert_newlines: string -> string.  Convert a multi-line string from
# Perforce's format (\n only) to TeamTrack's format (\r\n).

def convert_newlines(s):
    # Remove final newline (if any).
    if s and s[-1] == '\n':
        s = s[:-1]
    # Replace \n with \r\n.
    s = string.replace(s, '\n', '\r\n')
    return s

# Connect to the Perforce server.
p4i = p4.p4({'p4-port': p4_port,
             'p4-user': p4_user,
             'p4-password': p4_password, })

# Connect to the TeamTrack server.
tt = teamtrack.connect(teamtrack_user,
                       teamtrack_password,
                       teamtrack_server)

# *** Fetch Perforce jobs for migrating.  To start with, I suggest only
# migrating a single job, to check that the script is working correctly.
# Specify the job you want to migrate in the -e option to the jobs command,
# like this:
jobs = p4i.run('jobs -e "job=job000001"')

# *** When you are happy that the migration is working correctly, fetch all the
# jobs, like this: jobs = p4i.run('jobs')

# Now submit each Perforce job as a TeamTrack case.
for job in jobs:
    # Make a TeamTrack case that will correspond to the job.
    case = tt.new_record(teamtrack.table['CASES'])

    # Now fill in each of the case's fields with appropriate values from the
    # job.  Five fields are normally required by TeamTrack: TITLE, DESCRIPTION,
    # ISSUETYPE, PROJECTID and SEVERITY.

    # 1. *** The TeamTrack TITLE should be a one-sentence summary and the
    # DESCRIPTION a detailed explanation.  If you've been using the default
    # Perforce jobspec, then you only have one field, the Description.  So I
    # suggest making the TITLE be the first line of the description, and the
    # DESCRIPTION be the full description.  You may need to go into TeamTrack
    # and fix things up later by hand.
    case['TITLE'] = job['Description'][0:string.find(job['Description'],'\n')]
    case['DESCRIPTION'] = convert_newlines(job['Description'])

    # 2. *** The ISSUETYPE indicates whether the case is a bug or change
    # request or whatever.  The legal values for this field are found in the
    # TS_ID field in the TS_SELECTIONS table where TS_FLDID = 4.  I have chosen
    # the value 1, which means "Bug Report" in my TeamTrack database (you must
    # check yours to find an appropriate value).
    case['ISSUETYPE'] = 1

    # 3. *** The legal values for the PROJECTID field are found in the TS_ID
    # field in the TS_PROJECTS table.  Project 2 is the first project in my
    # TeamTrack database (you must check yours to find an appropriate value).
    # Or you could compute the TeamTrack project based on the Perforce project
    # if your Perforce jobs have a project field.
    case['PROJECTID'] = 2
    
    # 4. *** The legal values for the SEVERITY field are found in in the TS_ID
    # field in the TS_SELECTIONS table where TS_FLDID = 14.  I have chosen the
    # value 30, which means "Medium" in my TeamTrack database (you must check
    # yours to find an appropriate value).  Better still would be to compute
    # the TeamTrack severity based on the Perforce severity (if your Perforce
    # jobs have a severity field).
    case['SEVERITY'] = 30

    # 5. *** You don't have to fill in any more fields in the case if you don't
    # want to.  But if you have more fields in Perforce, you'll probably want
    # to fill in some more fields in the TeamTrack case.  You will need to read
    # the TeamTrack database schema [TeamShare 2000-01-20] to understand the
    # legal values for fields in a TeamTrack case.  The two fields you can't
    # set at this point are the STATE field (because when the case is submitted
    # it ends up in state New or whatever you've defined to be the first state
    # in the workflow) and the OWNER field (because TeamTrack automatically
    # sets the OWNER field when you submit an issue); see steps 11 and 12
    # below.

    # 6. Set P4DTI fields based on the configuration.
    case['P4DTI_RID'] = rid
    case['P4DTI_SID'] = sid
    case['P4DTI_ACTION'] = 'replicate'

    # 7. *** Set the jobname that the case will be replicated to.  If you want
    # to keep jobnames the same in Perforce, then use the jobname here.  If you
    # would prefer the Perforce jobname to match the case id, comment out the
    # next line.
    case['P4DTI_JOBNAME'] = job['Job']

    # 8. *** Pick a user on whose behalf to submit the case.  If your Perforce
    # userids are the same as the TeamTrack userids, then it should be safe to
    # use the User field of the job.  Otherwise, you'll have to convert the
    # Perforce user somehow.
    user = job['User']

    # 9. Submit the case.
    issueid = case.submit(user)

    # 10. Fetch the case from the database so that we can set the STATE and
    # OWNER fields.
    case = tt.query(teamtrack.table['CASES'],
                    "TS_ISSUETYPE = %d AND TS_ISSUEID = '%05d'"
                    % (case['ISSUETYPE'], issueid))[0]

    # 11. *** Set the STATE based on the Perforce job status.  The state_table
    # is a map from Perforce job status to TeamTrack state.  The legal values
    # for TeamTrack states are given by the TS_ID field in the TS_STATES table;
    # you must look in your TeamTrack database for the right values.
    state_table = { 'open': 2,
                    'closed': 3,
                    'suspended': 5, }
    case['STATE'] = state_table[job['Status']]

    # 12. Set the OWNER field to be the user who submitted the case (who came
    # from the job's user field in step 6 above).
    case['OWNER'] = case['SUBMITTER']

    # 13. Update the case in the database (note that by using 'update' rather
    # than 'transition' we bypass TeamTrack's checks on the workflow).
    case.update()

    # 14. Report on what's been done.
    print ("Submitted a case with issueid %d corresponding to job %s"
           % (issueid, job['Job']))
    print '-' * 60
    print case
    print '-' * 60
    print job
    print '-' * 60



# A. References
#
# [TeamShare 2000-01-20] "TeamTrack Database Schema (Database Version: 21)";
# TeamShare; 2000-01-20; <URL: http://www.ravenbrook.com/project/p4dti/import/2000-01-20/teamtrack-schema/TTschema21.pdf>.
#
#
# B. Document History
# 
# 2000-10-19 GDR Created.
#
# 2000-12-08 RB  Moved config_teamtrack to init_teamtrack.
#
# 2000-12-18 NB  Moved init_teamtrack to init.
#
# 2001-02-22 GDR Rewritten so it can run before you've installed the P4DTI.
#
# 2001-02-23 GDR Added STATE and OWNER.  Set the P4DTI fields, including
# P4DTI_JOBNAME.
#
# 2001-03-02  RB    Transferred copyright to Perforce under their license.
