# p4.py -- Simple Python interface to Perforce, using p4 -G
# Gareth Rees, Ravenbrook Limited, 2000-09-25.
# $Id: //info.ravenbrook.com/project/p4dti/version/1.0/code/replicator/p4.py#4 $
#
# This file is copyright (c) 2001 Perforce Software, Inc.  All rights  
# reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1.  Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
# 
# 2.  Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDERS AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

import marshal
import os
import re
import tempfile
import types

error = 'Perforce error'

class p4:
    config = { 'p4-client': None,
               'p4-client-executable': 'p4',
               'p4-password': None,
               'p4-port': None,
               'p4-user': None,
               'logger': None,
               'verbose': 0,
               }

    def __init__(self, config = {}):
        # Merge the supplied config with the default config (the former takes
        # precedence).
        for k in config.keys():
            self.config[k] = config[k]

        # Check that the Perforce client version supports "p4 -G".
        changelevel = 0
        supported_changelevel = 16895
        p4_exe = self.config['p4-client-executable']
        if ' ' in p4_exe:
            command = '"%s" -V' % p4_exe
        else:
            command = '%s -V' % p4_exe
        lines = os.popen(command, 'r').readlines()
        client_version_re = re.compile('Rev\\. [^/]+/[^/]+/[^/]+/([0-9]+)')
        for l in lines:
            match = client_version_re.match(l)
            if match:
                changelevel = int(match.group(1))
                if changelevel < supported_changelevel:
                    raise error, \
                          ("Perforce client changelevel %d is not "
                           "supported by P4DTI. client must be at "
                           "changelevel %d or above"
                           % (changelevel, supported_changelevel))
                else:
                    break
        if not changelevel:
            raise error, ("The command '%s' didn't report a recognizable "
                          "version number.  Check your setting for the "
                          "p4_client_executable parameter." % command)
        

    def log(self, format, arguments = (), priority = None):
        if self.config['logger']:
            format = "%s\t" + format
            if isinstance(arguments, types.TupleType):
                arguments = ("p4") + arguments
            else: # single argument
                arguments = ("p4", arguments)
            self.config['logger'].log("P4DTI-0",
                                      format, arguments, priority)


    # p4.run(arguments, input, verbose).  Run the p4 client with the
    # given arguments.  The arguments should be a Perforce command and
    # its arguments, like "jobs -o //foo/...".  Options should
    # generally include -i and/or -o to avoid forms being put up
    # interactively.  If input is supplied, then it should be a list
    # of dictionaries.  If verbose is true, then the input, command
    # and results will be printed.  These dictionaries are sent one by
    # one to the process.  The results are read into a list and
    # returned.  Since the output of the process is read to EOF and
    # the input is closed, there should be no process left hanging
    # about.

    def run(self, arguments, input = None, verbose = 0):
        temp_filename = None
        # Quote the Perforce command if it contains spaces.  See job000049.
        p4_exe = self.config['p4-client-executable']
        if ' ' in p4_exe:
            command_words = ['"%s"' % p4_exe]
        else:
            command_words = [p4_exe]
        command_words.append('-G')
        if self.config['p4-port']:
            command_words.append('-p %s' % self.config['p4-port'])
        if self.config['p4-user']:
            command_words.append('-u %s' % self.config['p4-user'])
        if self.config['p4-password']:
            command_words.append('-P %s' % self.config['p4-password'])
        if self.config['p4-client']:
            command_words.append('-c %s' % self.config['p4-client'])
        command_words.append(arguments)
        if input:
            tempfile.template = 'p4dti_data'
            temp_filename = tempfile.mktemp()
            temp_file = open(temp_filename, 'wb')
            marshal.dump(input[0], temp_file)
            temp_file.close()
            command_words.extend(['<', temp_filename])
            if verbose or self.config['verbose']:
                self.log ("p4 input: '%s'", str(input))
        command = reduce((lambda x,y: x + ' ' + y), command_words)
        if verbose or self.config['verbose']:
            self.log ("p4 command: '%s'", command)
            
        assert (os.name == 'nt' or os.name == 'posix'), \
               ("p4.run does not support os.name='%s'." % os.name)
        if os.name == 'nt' :
            mode = 'rb'
        elif os.name == 'posix' :
            mode = 'r'
        stream = os.popen(command, mode)

        results = []
        try:
            while 1:
                results.append(marshal.load(stream))
        except EOFError:
            if (temp_filename):
                os.remove(temp_filename)
        # Check the exit status of the Perforce command, rather than
        # simply returning empty output.  This code was inserted to
        # resolve job job000158.  RB 2000-12-14
        exit_status = stream.close()
        if verbose or self.config['verbose']:
            self.log("p4 status: '%s'", exit_status)
            self.log("p4 results: '%s'", str(results))
        # Note: this isn't a totally reliable way to spot an error in
        # a Perforce command.  See job000003.
        if (len(results) == 1 and results[0].has_key('code')
            and results[0]['code'] == 'error'):
            if exit_status:
                raise error, ("%s  The Perforce client exited with error "
                              "code %d." % (results[0]['data'], exit_status))
            else:
                raise error, results[0]['data']
        elif exit_status:
            raise error, \
                  ("The Perforce client exited with error code %d.  "
                   "Is the server down or the server address incorrect?" %
                   exit_status)
        else:
            return results


# B. Document History
# 
# 2000-12-07 GDR Provided defaults for all configuration parameters so that you
# can make a p4 object passing no parameters to get the default Perforce
# behaviour.
#
# 2000-12-14 RB  Added check for the exit status of the "p4" command so that
# the caller can tell the difference between empty output and a connection
# (or other) error.
#
# 2000-12-15 NB  Added verbosity control.
#
# 2001-01-23 GDR Added check that Perforce client version is supported.
#
# 2001-02-14 GDR Report the Perforce error message together with the exit
# status when we have both.
#
# 2001-02-19 NB Keyword translation updated and moved here (as it is
# Perforce-specific.
#
# 2001-02-21 GDR Moved keyword translator to its own file (keyword.py) so that
# there's no import loop.
#
# 2001-03-02  RB    Transferred copyright to Perforce under their license.
